from __future__ import annotations
from pathlib import Path

from whyqd.models import SchemaModel, CrosswalkModel
from whyqd.core import SchemaDefinition
from whyqd.crud import CRUDAction
from whyqd.core.base import BaseDefinition


class CrosswalkDefinition(BaseDefinition):
    """Create and manage a method to perform a schema to schema crosswalk.

    !!! tip "Strategy"
        Guidance on how to use this definition is in the strategies section on
        [Crosswalk Strategies](../strategies/crosswalk.md).

    Parameters:
      crosswalk: A dictionary conforming to the CrosswalkModel, or a path to a saved definition.
      schema_source: Path to a json file containing a saved schema, or a dictionary conforming to one, or a definition.
      schema_destination: Path to a json file containing a saved schema, or a dictionary conforming to one, or a definition.

    Example:
      Create a new `CrosswalkDefinition` as follows:

      ```python
      import whyqd as qd

      crosswalk = qd.CrosswalkDefinition()
      crosswalk.set(schema_source=SCHEMA_SOURCE, schema_destination=SCHEMA_DESTINATION)
      ```
    """

    def __init__(
        self,
        *,
        crosswalk: CrosswalkModel | dict | Path | str | None = None,
        schema_source: SchemaDefinition | SchemaModel | dict | Path | str | None = None,
        schema_destination: SchemaDefinition | SchemaModel | dict | Path | str | None = None,
    ) -> None:
        super().__init__(model_name="crosswalk")
        self.crud = CRUDAction()
        self.schema_source = SchemaDefinition()
        self.schema_destination = SchemaDefinition()
        self.set(crosswalk=crosswalk, schema_source=schema_source, schema_destination=schema_destination)

    def _refresh_model_terms(self) -> None:
        """Refreshes the list of terms and connects them to the Form Model"""
        if self.model:
            self.model.actions = self.crud.get_all()

    def _refresh_model_fields(self) -> None:
        """Refreshes the list of terms and connects them to the Field CRUD"""
        if self.model and self.model.actions:
            self.crud.add_multi(terms=self.model.actions)
        if self.model and not self.model.actions:
            self.crud.reset()

    @property
    def get(self) -> CrosswalkModel | None:
        """Get the crosswalk model.

        Returns:
          A Pydantic CrosswalkModel or None
        """
        self._refresh_model_terms()
        return self.model

    def set(
        self,
        *,
        crosswalk: CrosswalkModel | dict | Path | str | None = None,
        schema_source: SchemaDefinition | SchemaModel | dict | Path | str | None = None,
        schema_destination: SchemaDefinition | SchemaModel | dict | Path | str | None = None,
    ) -> None:
        """Update or create the crosswalk.

        Parameters:
          crosswalk: A dictionary conforming to the CrosswalkModel, or a path to a saved definition.
          schema_source: Path to a json file containing a saved schema, or a dictionary conforming to one, or a definition.
          schema_destination: Path to a json file containing a saved schema, or a dictionary conforming to one, or a definition.
        """
        self.model = self.core.create_or_update_model(modelType=CrosswalkModel, source=crosswalk, model=self.model)
        # And update the original data
        # https://fastapi.tiangolo.com/tutorial/body-updates/#partial-updates-with-patch
        if self.model.schemaSource and not schema_source:
            self.schema_source.set(schema=self.model.schemaSource)
        elif schema_source:
            if isinstance(schema_source, SchemaDefinition):
                self.schema_source = schema_source
            else:
                self.schema_source.set(schema=schema_source)
            self.model.schemaSource = self.schema_source.get
        if self.model.schemaDestination and not schema_destination:
            self.schema_destination.set(schema=self.model.schemaDestination)
        elif schema_destination:
            if isinstance(schema_destination, SchemaDefinition):
                self.schema_destination = schema_destination
            else:
                self.schema_destination.set(schema=schema_destination)
            self.model.schemaDestination = self.schema_destination.get
        if self.model.schemaSource and self.model.schemaDestination:
            self.crud.set_schema(schema_source=self.schema_source, schema_destination=self.schema_destination)
            self._refresh_model_fields()

    #########################################################################################
    # MANAGE ACTION SCRIPTS
    #########################################################################################

    @property
    def actions(self) -> CRUDAction:
        """Returns the active crud model for all Action operations. See [Action CRUD](action.md) for API.

        Returns:
          For all Action CRUD behaviours.
        """
        if not self.model.schemaSource or not self.model.schemaDestination:
            raise ValueError("First set source and destination schemas before performing defining a crosswalk.")
        return self.crud

    #########################################################################################
    # VALIDATION UTILITIES
    #########################################################################################

    def validate(self) -> bool:
        """Validate that all required fields are returned from the crosswalk.

        Raises:
          ValueError: If required destination fields are not present in the crosswalk.

        Returns:
          A boolean `True` if successful.
        """
        required = self.crud.validate(required=True)
        if required:
            raise ValueError(
                f"Validation error. Required destination fields are not present in the crosswalk: {required}"
            )



from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional
from uuid import UUID, uuid4

from whyqd.models.constraints import ConstraintsModel
from whyqd.dtypes import FieldType


class FieldModel(BaseModel):
    uuid: UUID = Field(default_factory=uuid4, description="Unique identity for the field. Automatically generated.")
    name: str = Field(
        ...,
        description="Defined exactly as it appears in the source data. By convention, it should be a machine-readable term to uniquely address this field. Should not have spaces. CamelCase or snake_case for preference.",
    )
    title: Optional[str] = Field(None, description="A human-readable version of the field name.")
    description: Optional[str] = Field(None, description="A description for the field.")
    dtype: FieldType = Field(..., alias="type", description="A field must contain values of a specific type.")
    example: Optional[str] = Field(None, description="An example value, as a string, for the field.")
    constraints: Optional[ConstraintsModel] = Field(
        None, description="A set of optional constraints to define the field."
    )

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True
        allow_population_by_field_name = True



from __future__ import annotations
from typing import Generic, TypeVar, Type
from pydantic import BaseModel
from uuid import UUID

ModelType = TypeVar("ModelType", bound=BaseModel)


class CRUDBase(Generic[ModelType]):
    """
    CRUD object with default methods to Create, Read, Update, Delete (CRUD).

    Parameters:
      model: A Pydantic model class
    """

    def __init__(self, model: Type[ModelType]):
        self.model = model
        self.multi = []

    def get(self, *, name: str | UUID) -> ModelType | None:
        """Get a specific model from the list of models defining this schema, called by a unique `name`.

        Parameters:
          name: Specific name or reference UUID for a field already in the Schema.

        Returns:
          ModelType, or None if no such `name` or `UUID`.
        """
        name = self.get_hex(name=name)
        if self.multi:
            # https://stackoverflow.com/a/31988734/295606
            # It is statistically almost impossible to have a field name that matches a randomly-generated UUID
            # Can be used to recover fields from hex
            return next((m for m in self.multi if m.name == name or m.uuid.hex == name), None)
        return None

    def get_all(self) -> list[ModelType]:
        """Get all models from the current list of models."""
        return self.multi

    def add(self, *, term: ModelType | dict) -> None:
        """Add the parameters for a specific term, called by a unique `name`. If the `name` already exists, then this
        will raise a `ValueError`.

        Parameters:
          term: A dictionary conforming to the ModelType. Model names must be unique, so a valid `name` in the model list
                will have no collisions.

        Raises:
          ValueError: If the term already exists.
        """
        if isinstance(term, dict):
            term = self.model(**term)
        if self.get(name=term.name):
            raise ValueError(f"ModelType {term.name} already exists.")
        self.multi.append(term)

    def add_multi(self, *, terms: list[ModelType | dict]) -> None:
        """Add multiple parameters for a specific term, called by a unique `name`. If the `name` already exists, then
        this will raise a `ValueError`.

        Parameters:
          terms: A list of dictionaries conforming to the ModelType.

        Raises:
          ValueError: If the term already exists.
        """
        for term in terms:
            self.add(term=term)

    def update(self, *, term: ModelType | dict) -> None:
        """Update the parameters for a specific term, called by a unique `name`. If the `name` does not exist, then
        this will raise a `ValueError`.

        Parameters:
          term: A dictionary conforming to the ModelType.

        Raises:
          ValueError: If the term does not exist
        """
        if isinstance(term, dict):
            term = self.model(**term)
        # Create a temporary FieldModel
        old_term = self.get(name=term.name)
        if not old_term:
            raise ValueError(f"ModelType {term.name} does not exist.")
        # And update the original data
        old_name = old_term.name
        # https://fastapi.tiangolo.com/tutorial/body-updates/#partial-updates-with-patch
        old_term = old_term.copy(update=term.dict(exclude_unset=True))
        self.remove(name=old_name)
        self.add(term=old_term)

    def remove(self, *, name: str) -> None:
        """Remove a specific term, called by a unique `name`.

        Parameters:
          name: Specific name or reference UUID for a field already in the Schema.
        """
        # https://stackoverflow.com/a/1235631/295606
        name = self.get_hex(name=name)
        self.multi[:] = [m for m in self.multi if m.name != name and m.uuid.hex != name]

    def reset(self) -> None:
        """Reset a list of ModelType terms to an empty list."""
        self.multi = []

    def reorder(self, *, order: list[UUID]) -> None:
        """Reorder a list of terms.

        Parameters:
          order: list of UUID in the desired order. Use `.get_all()` to view the current list.

        Raises:
          ValueError: If the list of `UUIDs` doesn't conform to that in the list of terms.
        """
        order = [self.get_hex(name=o) for o in order]
        if {m.uuid.hex for m in self.multi}.difference(set(order)):
            raise ValueError("List of reordered term ids isn't the same as that in the provided list of Models.")
        if self.multi:
            self.multi = sorted(self.multi, key=lambda term: order.index(term.uuid.hex))

    def get_hex(self, *, name: str | UUID) -> str:
        if isinstance(name, UUID):
            return name.hex
        return name



from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional
from datetime import datetime


class VersionModel(BaseModel):
    name: Optional[str] = Field(None, description="Name of the person who modified the schema or method.")
    description: Optional[str] = Field(None, description="Description of what was done in this version.")
    updated: Optional[datetime] = Field(
        default_factory=datetime.utcnow,
        description="Date generated on update.",
    )



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseMorphAction

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseMorphAction):
    """Delete rows provided in a list. They don't have to be contiguous.

    !!! tip "Script template"
        ```python
        "DELETE_ROWS < [int, int, int, etc.]"
        ```

        Where `int` are  specific integer row references / indices. Delete specifically does *not* reindex the rows so
        that future reference calls will reference a static index.

    !!! example
        ```python
        "DELETE_ROWS < [0, 1, 2, 3]"
        ```
    """

    def __init__(self) -> None:
        self.name = "DELETE_ROWS"
        self.title = "Delete rows"
        self.description = "Delete rows provided in a list. They don't have to be contiguous."
        self.structure = ["rows"]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        rows: list[int] = None,
        destination: None = None,
        source: None = None,
        source_param: None = None,
        destination_param: None = None,
    ) -> pd.DataFrame:
        return df.drop(df.index.intersection(rows))



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseMorphAction
from whyqd.models import FieldModel

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseMorphAction):
    """
    Unite a list of columns with a selected `string` (i.e. concatenating text in multiple fields).

    !!! tip "Script template"
        ```python
        "UNITE > 'destination_field' < 'by'::['source_field', 'source_field', etc.]"
        ```

        Or:
        ```python
        "UNITE > 'destination_field' < ['source_field', 'source_field', etc.]"
        ```

        Where `'by'` can be any `string` with one **MAJOR** caveat. Since `::` is used in the query, you **cannot** use it
        for joining. By default this will be `', '` and so, alternatively, you can leave the `'by'` term out.

    !!! example
        ```python
        "UNITE > 'reference' < ', '::['Reference 1', 'Reference 2', 'Reference 3']"
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "UNITE"
        self.title = "Unite"
        self.description = "Unite values in different fields to create a new concatenated value. Each value will be converted to a string (e.g. A: 'Word 1' B: 'Word 2' => 'Word 1 Word 2')."
        self.structure = ["fields"]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel | list[FieldModel],
        source: list[FieldModel],
        source_param: str | None = None,
        rows: None = None,
        destination_param: None = None,
    ) -> pd.DataFrame:
        if isinstance(destination, list):
            destination = destination[0]
        if source_param is None:
            source_param = ", "
        fields = [field.name for field in source]
        # https://stackoverflow.com/a/45976632
        df[destination.name] = df[fields].apply(
            lambda x: "" if x.isnull().all() else source_param.join(x.dropna().astype(str)).strip(), axis=1
        )
        return df



from __future__ import annotations
from typing import TYPE_CHECKING
from uuid import UUID
import modin.pandas as pd

from whyqd.crud.base import CRUDBase
from whyqd.models import ActionScriptModel
from whyqd.parsers import ScriptParser, ActionParser, MorphParser, CategoryParser
from whyqd.crosswalk.base import BaseSchemaAction, BaseMorphAction, BaseCategoryAction

if TYPE_CHECKING:
    from whyqd.core import SchemaDefinition
    from whyqd.models import FieldModel


class CRUDAction(CRUDBase[ActionScriptModel]):
    """Create, Read, Update and Delete Field Models. Usually instantiated as part of a
    [CrosswalkDefinition](crosswalk.md) and accessed as `.actions`.

    [Base CRUD operations](basecrud.md) are common for both `CRUDField` and `CRUDAction`.

    Example:
      ```python
      import whyqd as qd

      crosswalk = qd.CrosswalkDefinition()
      crosswalk.set(schema_source=SCHEMA_SOURCE, schema_destination=SCHEMA_DESTINATION)
      # Create the crosswalk
      schema_scripts = [
          "DEBLANK",
          "DEDUPE",
          "DELETE_ROWS < [0, 1, 2, 3]",
          f"PIVOT_LONGER > ['year', 'values'] < {datasource.model[0].names[4:]}",
          "RENAME > 'indicator_code' < ['Indicator Code']",
          "RENAME > 'indicator_name' < ['Indicator Name']",
          "RENAME > 'country_code' < ['Country Code']",
          "RENAME > 'country_name' < ['Country Name']",
      ]
      crosswalk.actions.add_multi(terms=schema_scripts)
      ```

    Parameters:
      schema_source: A [SchemaDefinition](schema.md) for access to source schema definitions and operations.
      schema_destination: A [SchemaDefinition](schema.md) for access to destination schema definitions and operations.
    """

    def __init__(self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None):
        super().__init__(ActionScriptModel)
        self.parser = ScriptParser()
        self.schema_source = None
        self.schema_destination = None
        self.uncrossed = []
        if schema_source and schema_destination:
            self.set_schema(schema_source=schema_source, schema_destination=schema_destination)

    def get(self, *, name: str | UUID) -> ActionScriptModel | None:
        """Get a specific script from the list of scripts, called by a unique id.

        Parameters:
          name: Scripts may be duplicated, but UUIDs will be unique.

        Returns:
          An ActionScriptModel or None, of no such script is found.
        """
        uid = self.get_hex(name=name)
        if self.multi:
            return next((m for m in self.multi if m.uuid.hex == uid), None)
        return None

    def add(self, *, term: str | ActionScriptModel) -> None:
        """Add the string term for an action script. Validate as well. Does not test for uniqueness.

        Parameters:
          term: A string term, or ActionScriptModel conforming to the script structure for a specific action.

        Raises:
          ValueError: If script is invalid. In most cases, you should get reasonable feedback from the error message
                      as to how to fix the problem.
        """
        # Validate the script

        if isinstance(term, (dict, ActionScriptModel)):
            if isinstance(term, dict):
                term = ActionScriptModel(**term)
            parsed = self.parse(script=term.script)
            self.multi.append(term)
        else:
            parsed = self.parse(script=term)
            self.multi.append(ActionScriptModel(**{"script": term}))
        self.reconcile_crosswalk(fields=parsed.get("destination", []))

    def update(self, *, term: ActionScriptModel | dict) -> None:
        """Update the parameters for a specific term, called by a unique `UUID`. If the `UUID` does not exist, then this
        will raise a `ValueError`.

        Parameters:
          term: A dictionary conforming to the ActionScriptModel.

        Raises:
          ValueError: If the term does not exist.
        """
        if isinstance(term, dict):
            term = self.model(**term)
        # Create a temporary FieldModel
        old_term = self.get(name=term.uuid.hex)
        if not old_term:
            raise ValueError(f"ActionScriptModel {term.script} does not exist.")
        # And update the original data
        parsed = self.parse(script=term)
        # It parses, ensure reconciliation
        old_parsed = self.parse(script=old_term)
        self.reconcile_crosswalk(fields=old_parsed.get("destination", []), remove=True)
        old_term.script = term.script
        self.reconcile_crosswalk(fields=parsed.get("destination", []))

    def remove(self, *, name: str | UUID) -> None:
        """Remove a specific term, called by a unique `UUID`.

        Parameters:
          name: Unique model UUID for the script.
        """
        script = self.get(name=name)
        if script:
            parsed = self.parse(script=script)
            # https://stackoverflow.com/a/1235631/295606
            self.multi[:] = [m for m in self.multi if m.uuid.hex != script.uuid.hex]
            self.reconcile_crosswalk(fields=parsed.get("destination", []), remove=True)

    def parse(self, *, script: str | ActionScriptModel) -> dict:
        """Parse a script for any action, of any type, and return the corresponding transformation dictionary structure.
        Can also be used as a validation step.

        Parameters:
          script: A string term, or ActionScriptModel conforming to the script structure for a specific action.

        Raises:
          ValueError: If the script term is not valid, or if the schemas are not set.

        Returns:
         ```python
          # A dictionary of the appropriate form for the ACTION
          {
              "action": BaseSchemaAction,
              "destination": FieldModel,
              "source": list[ModifierModel | FieldModel]
          }

          {
              "action": BaseMorphAction,
              "destination": FieldModel | list[FieldModel],
              "source": list[FieldModel],
              "rows": list[int],
              "source_param": str,
              "destination_param": str
          }

          {
              "action": BaseCategoryAction,
              "destination": FieldModel,
              "category": CategoryModel,
              "source": list[FieldModel],
              "assigned": list[CategoryModel],
              "unassigned": list[CategoryModel],
          }
          ```
        """
        if isinstance(script, ActionScriptModel):
            script = script.script
        action = self.get_action(script=script)
        action_parser = self.get_action_parser(script=script, action=action)()
        action_parser.set_schema(schema_source=self.schema_source, schema_destination=self.schema_destination)
        return action_parser.parse(script=script, action=action)

    def validate(self, *, required: bool = False) -> list[FieldModel]:
        """Return the list of destination schema fields which are still to be crosswalked.

        Parameters:
          required: Limit the unreconciled crosswalk fields to only those which are required.

        Returns:
          List of FieldModels which are required but are still undefined by an ActionScriptModel.
        """
        if required:
            return [
                f
                for f in self.schema_destination.get.fields
                if f.uuid in self.uncrossed and f.constraints and f.constraints.required
            ]
        return [f for f in self.schema_destination.get.fields if f.uuid in self.uncrossed]

    ###################################################################################################
    ### TABULAR DATA TRANSFORMATION UTILITIES
    ###################################################################################################

    def transform(self, *, df: pd.DataFrame, script: str | ActionScriptModel) -> pd.DataFrame:
        """Return a transformed dataframe according to a single script.

        Parameters:
          df: A Pandas (Modin) dataframe.
          script: A string term, or ActionScriptModel conforming to the script structure for a specific action.

        Returns:
          Transformed dataframe.
        """
        parsed = self.parse(script=script)
        action_parser = self.get_action_parser(script=script, action=parsed["action"])()
        action_parser.set_schema(schema_source=self.schema_source, schema_destination=self.schema_destination)
        return action_parser.transform(df=df, **parsed)

    def transform_all(self, *, df: pd.DataFrame) -> pd.DataFrame:
        """Return a transformed dataframe after processing all scripts.

        Parameters:
          df: A Pandas (Modin) dataframe.

        Returns:
          Transformed dataframe.
        """
        for script in self.get_all():
            df = self.transform(df=df, script=script)
        return df

    ###################################################################################################
    ### ACTION SUPPORT UTILITIES
    ###################################################################################################

    def set_schema(
        self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None
    ) -> None:
        """Set SchemaDefinitions.

        Parameters:
          schema_source: A [SchemaDefinition](schema.md) for access to source schema definitions and operations.
          schema_destination: A [SchemaDefinition](schema.md) for access to destination schema definitions and operations.
        """
        if not schema_source or not schema_destination:
            raise ValueError("Schema for both source and destination has not been provided.")
        self.schema_source = schema_source
        self.schema_destination = schema_destination
        self.uncrossed = {field.uuid for field in self.schema_destination.get.fields}

    def get_action(self, *, script: str) -> BaseSchemaAction | BaseMorphAction | BaseCategoryAction:
        """Return the first action term from a script as its Model type.

        Parameters:
          script: A string term conforming to the script structure for a specific action.

        Raises:
          ValueError: If not a valid ACTION.

        Returns:
          The action model type conforming to the requirements for a script.
        """
        action = self.parser.get_action_from_script(script=script)
        return self.parser.get_action_class(actn=action)()

    def get_action_parser(
        self, *, script: str, action: BaseSchemaAction | BaseMorphAction | BaseCategoryAction = None
    ) -> ActionParser | MorphParser | CategoryParser:
        """Return the ACTION parser for a script.

        Parameters:
          script: A string term conforming to the script structure for a specific action.
          action: The action model type conforming to the requirements for a script.

        Raises:
          ValueError: If the script can't be parsed.

        Returns:
          Parser for Action type, and used in [TransformDefinition](transform.md) and [CrosswalkDefinition](crosswalk.md).
        """
        if not action:
            action = self.get_action(script=script)
        # Test for each of BaseSchemaAction | BaseMorphAction | BaseCategoryAction
        if isinstance(action, BaseSchemaAction):
            return ActionParser
        if isinstance(action, BaseMorphAction):
            return MorphParser
        if isinstance(action, BaseCategoryAction):
            return CategoryParser
        # Action unknown
        raise ValueError(f"Script cannot be parsed ({script}).")

    def reconcile_crosswalk(self, *, fields: FieldModel | list[FieldModel], remove: bool = False) -> None:
        if fields is None:
            return
        if not isinstance(fields, list):
            fields = [fields]
        for f in fields:
            if remove:
                self.uncrossed = self.uncrossed ^ {f.uuid}
            else:
                self.uncrossed = self.uncrossed - {f.uuid}



from __future__ import annotations
from pydantic import BaseModel, Field, validator, constr
from typing import List, Optional, Union
from uuid import UUID, uuid4

from whyqd.models import FieldModel, VersionModel, CitationModel


class SchemaModel(BaseModel):
    uuid: UUID = Field(default_factory=uuid4, description="Automatically generated unique identity for the schema.")
    name: constr(strip_whitespace=True, to_lower=True) = Field(
        ...,
        description="Machine-readable term to uniquely address this schema. Cannot have spaces. CamelCase or snake_case for preference.",
    )
    title: Optional[str] = Field(None, description="A human-readable version of the schema name.")
    description: Optional[str] = Field(
        None,
        description="A complete description of the schema. Depending on how complex your work becomes, try and be as helpful as possible to 'future-you'. You'll thank yourself later.",
    )
    fields: List[FieldModel] = Field(
        default=[],
        description="A list of fields which define the schema. Fields, similarly, contain `name`, `title` and `description`, as well as `type` as compulsory.",
    )
    missingValues: list[str] = Field(
        default=[],
        description="Indicates which string values should be treated as null values. There could be a variety of these, such as '..', or '-'.",
    )
    primaryKey: Optional[Union[list[str], str]] = Field(
        default=None,
        description="A field or set of fields which uniquely identifies each row in the table. Specify using the `name` of relevant fields.",
    )
    index: Optional[int] = Field(
        None,
        description="Maximum value of a zero-base index for tabular data defined by this schema. Necessary where `actions` apply row-level transforms.",
    )
    citation: Optional[CitationModel] = Field(None, description="Optional full citation for the source data.")
    version: List[VersionModel] = Field(default=[], description="Version and update history for the schema.")

    class Config:
        # anystr_strip_whitespace = True
        validate_assignment = True

    @validator("name")
    def name_space(cls, v):
        return "_".join(v.split(" ")).lower()

    @validator("fields")
    def are_fields_unique(cls, v):
        field_names = []
        for f in v:
            field_names.append(f.name)
        if len(field_names) != len(set(field_names)):
            raise ValueError(
                f"Field names must be unique. There are {len(field_names) - len(set(field_names))} duplications."
            )
        return v



from __future__ import annotations
from typing import Union, Optional, Type, TYPE_CHECKING
from uuid import uuid4

from whyqd.parsers import CoreParser, ScriptParser

from whyqd.models import ConstraintsModel, FieldModel
from whyqd.crosswalk.base import BaseSchemaAction

if TYPE_CHECKING:
    from whyqd.models import SchemaActionModel, ModifierModel
    from whyqd.core import SchemaDefinition
    import modin.pandas as pd


class ActionParser:
    """Parsing functions for action scripts.

    Can process and validate any action script. Scripts may be recursive and are of the form:

        "ACTION > 'destination field' < [modifier 'source field', {action script}]"

    Modifiers are dependent on the specific Action. Fields must be found in provided SchemaDefinitions.
    """

    def __init__(self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None) -> None:
        """
        Parameters
        ----------
        schema_source: SchemaDefinition, optional
        schema_destination: SchemaDefinition, optional
        """
        self.core = CoreParser()
        self.parser = ScriptParser()
        self.schema = []
        if schema_source and schema_destination:
            self.set_schema(schema_source=schema_source, schema_destination=schema_destination)
        self.source_modifiers = {}
        self.modifier_names = set()

    ###################################################################################################
    ### PARSE TRANSFORM SCRIPT
    ###################################################################################################

    def parse(
        self, *, script: str, action: BaseSchemaAction
    ) -> dict[str, BaseSchemaAction | FieldModel | list[ModifierModel | FieldModel]]:
        """Parse a script for a base action and return the corresponding destination field, and source structure for
        transformation. Can, optionally, also update the existing schema list.

        Parameters
        ----------
        script: str
        action: BaseSchemaAction

        Raises
        ------
        ValueError if the script term is not recognised.

        Returns
        -------
        dict of the form:
            {
                'action': BaseSchemaAction,
                'destination': FieldModel | list[FieldModel],
                'source': list[ModifierModel | FieldModel]
            }
        """
        if not self.schema:
            raise ValueError("Schema has not been provided.")
        hexed_script = self.get_hexed_script(script=script)
        root = self.parser.get_split_terms(script=hexed_script, by="<")
        split_terms = self.parser.get_split_terms(script=root[0], by=">")
        # Get destination term
        destination = None
        if len(split_terms) == 2:
            destination = self.parser.get_literal(text=split_terms[1])
            if isinstance(destination, list):
                destination = [self.get_schema_field(term=d) for d in destination]
                assert all(isinstance(d, FieldModel) for d in destination)
            else:
                destination = self.get_schema_field(term=destination)
                assert isinstance(destination, FieldModel)
        # Get source term
        source = None
        if len(root) > 1:
            source = "<".join(root[1:])
        # Process initial response
        if action.settings.name == "NEW":
            # Special case where value is assigned as default to 'destination' field
            value = self.parser.get_literal(text=root[1])
            if len(value) > 1:
                raise ValueError(f"'New' actions must only contain a single value term. ({root[1:]})")
            destination.constraints = ConstraintsModel(**{"default": {"name": value[0]}})
            return {"action": action, "destination": destination}
        if not source:
            if action.settings.structure:
                # The structure for this action requires a source term
                raise ValueError(
                    f"{action.name} action requires a source term ({action.settings.structure}) but none found."
                )
            return {"action": action, "destination": destination}
        # If action does not include a structure, then no source term should be included
        if not action.settings.structure:
            # The structure for this action requires a source term
            raise ValueError(f"{action.settings.name} action does not include a source term but one found ({source}).")
        # Source exists *and* and is required, process the second part
        # Nested sources must not have destinations as these will be autogenerated
        last_i = None
        parsed = []
        for i, stack in list(self.parser.generate_contents(text=source)):
            # Generate contents yields the deepest, right-most nested bracket first, and goes from there.
            if not last_i:
                last_i = i
            if last_i == i:
                # Process nested sources on same level, e.g. [[nested_source1], [nested_source2]]
                parsed.append(
                    (stack, {"action": None, "source": self.parser.get_listed_literal(text=stack)}, uuid4().hex)
                )
            else:
                # Structure: ACTION < SOURCE
                # Process: as pop 'up' through nested sources, replace lower levels with hash key to simplify
                # extraction of the action, then replace the hash with a nested dictionary {action: , source: }
                # where each 'source' contains the input for that action, and is the input for the higher level.
                parsed_stack = stack
                for txt, prsed, hx in parsed:
                    # replace the txt with hx
                    parsed_stack = parsed_stack.replace(f"[{txt}]", hx)
                i_prsed = []
                for s in self.parser.get_split_terms(script=parsed_stack, by=",", maxsplit=-1):
                    splt = self.parser.get_split_terms(script=s, by="<")
                    if len(splt) == 1:
                        i_prsed.extend(self.parser.get_listed_literal(text=s))
                    else:
                        for txt, prsed, hx in parsed:
                            if hx in s:
                                prsed["action"] = self.parser.get_action_model(action=splt[0])
                                if not isinstance(prsed["action"], SchemaActionModel):
                                    raise ValueError(f"Only ACTIONS of Type `SchemaAction` can be nested ({stack}).")
                                i_prsed.append(prsed)
                parsed = [(stack, {"action": None, "source": i_prsed}, uuid4().hex)]
        # Once the stack is empty, need only the prsed 'source' section of the list
        source = [p[1]["source"] for p in parsed]
        source = self.recover_fields_from_hexed_script(parsed=source, action=action)[0]
        action.validate(destination=destination, source=source)
        return {"action": action, "destination": destination, "source": source}

    ###################################################################################################
    ### IMPLEMENT VALIDATED SCRIPT
    ###################################################################################################

    def transform(
        self,
        *,
        df: pd.DataFrame,
        action: Type[BaseSchemaAction],
        destination: FieldModel,
        source: Optional[list[Union[ModifierModel, dict]]] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        A recursive transformation. A method should be a list fields upon which actions are applied, but
        each field may have nested sub-fields requiring their own actions. Before the action on the
        current field can be completed, it is necessary to perform the actions on each sub-field.

        Parameters
        ----------
        df: DataFrame
            Working data to be transformed
        action: SchemaActionModel
        destination: FieldModel
        source: list of ModifierModel, and dicts of nested transforms, default None
        assigned: list of dict
            Specific to CATEGORISE actions. Each dict has values for: Assignment ACTION, destination schema field,
            schema category, source data column, and a list of source data column category terms assigned to that
            schema category.

        Returns
        -------
        Dataframe
            Containing the implementation of all nested transformations
        """
        if not source:
            return action.transform(df=df, destination=destination)
        flattened_source = []
        for term in source:
            if isinstance(term, dict):
                # Nested transform
                nested_destination = term.get("destination")
                if not nested_destination:
                    # Need to create a temporary column ... the action will be performed here
                    # then this nested structure will be replaced by the output of this new column
                    # Temporary column is based on initial root FieldModel destination to ensure
                    # all schema parameters carried over.
                    nested_destination = destination.copy()
                    nested_destination.name = f"nested_{uuid4().hex}"
                df = self.transform(df, term["action"], nested_destination, term.get("source"))
                flattened_source.append(nested_destination)
            else:
                flattened_source.append(term)
        # Action transform
        return action.transform(df=df, destination=destination, source=flattened_source)

    ###################################################################################################
    ### SUPPORT UTILITIES
    ###################################################################################################

    def set_schema(
        self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None
    ) -> None:
        """Set SchemaDefinitions for the parser.

        Parameters
        ----------
        schema_source: SchemaDefinition
        schema_destination: SchemaDefinition
        """
        if not schema_source or not schema_destination:
            raise ValueError("Schema for both source and destination has not been provided.")
        self.schema = [schema_source, schema_destination]

    def get_schema_field(self, *, term: str) -> FieldModel:
        """Recover a field model from a string.

        Raises
        ------
        ValueError if the field term is not recognised.

        Returns
        -------
        FieldModel
        """
        field = None
        for s in self.schema:
            field = s.fields.get(name=term)
            if field:
                break
        if not field:
            e = f"Field name is not recognised from either of the source or destination schema fields ({term})."
            if len(self.schema) == 1:
                e = f"Field name is not recognised from the schema fields ({term})."
            raise ValueError(e)
        return field

    def get_hexed_script(self, *, script: str) -> str:
        from whyqd.crosswalk.actions import default_actions

        # Changes fields to uuid hexes
        all_fields = [field for s in self.schema for field in s.get.fields]
        script = self.parser.get_hexed_script(script=script, fields=all_fields)
        self.modifier_names = set()
        self.source_modifiers = {}
        for action in default_actions:
            if action.name in script:
                action_modifiers = {}
                if action.modifiers:
                    # TODO: this isn't right ... modifiers may share a name but have different functions/descriptions
                    action_modifiers = {m.name: m for m in action.modifiers}
                for m in set(action_modifiers.keys()).difference(self.modifier_names):
                    # Preserve original Modifiers
                    self.source_modifiers[m] = action_modifiers[m]
                    script = script.replace(m, f",{m},")
                self.modifier_names.update(set(action_modifiers.keys()))
        return ",".join([s.strip() for s in script.split(",") if s.strip()])

    def recover_fields_from_hexed_script(
        self, *, parsed: list | dict, action: BaseSchemaAction
    ) -> list[dict[str, BaseSchemaAction | FieldModel | list[ModifierModel | FieldModel]]]:
        """
        Recovered fields and modifiers from a hexed script. Doubles up as a validator since any non-recovered
        term causes a ValueError. Recursive for deeply-nested scripts.

        Parameters
        ----------
        parsed: list, dict
        action: BaseSchemaAction

        Raises
        ------
        ValueError if the script term is not recognised.

        Returns
        -------
        list of dicts of the form:
            {
                'action': BaseSchemaAction,
                'destination': FieldModel,
                'source': list[ModifierModel | FieldModel]
            }
        """
        recovered_fields = []
        modifier_names = action.modifier_terms
        if not isinstance(parsed, list):
            parsed = [parsed]
        for term in parsed:
            if not term:
                # Blank string artifacts can be introduced
                continue
            recovered = None
            if isinstance(term, str) and term in modifier_names:
                recovered = action.get_modifier(term=term)
            elif isinstance(term, str):
                recovered = self.get_schema_field(term=term)
            elif isinstance(term, list):
                recovered = self.recover_fields_from_hexed_script(parsed=term, action=action)
            elif isinstance(term, dict):
                # Check if 'destination' is a column or schema field
                destination = term.get("destination")
                if destination:
                    destination = self.get_schema_field(term=destination)
                recovered = {
                    "action": term.get("action"),
                    "destination": destination,
                    "source": self.recover_fields_from_hexed_script(source=term.get("source"), action=action),
                }
            if not recovered:
                raise ValueError(f"Term ({term}) cannot be parsed.")
            recovered_fields.append(recovered)
        return recovered_fields



from __future__ import annotations
from typing import List, Optional, Union
from pydantic import BaseModel, Field, validator

from whyqd.models import CategoryModel


class ConstraintsModel(BaseModel):
    required: Optional[bool] = Field(
        default=False, description="Indicates whether a field must have a value for each instance."
    )
    unique: Optional[bool] = Field(
        default=False, description="When `true`, each value defined by the field `MUST` be unique."
    )
    category: Optional[List[CategoryModel]] = Field(
        None,
        alias="enum",
        description="The set of unique category terms permitted in this field,  with `name` & (optional) `description`.",
    )
    default: Optional[CategoryModel] = Field(
        None,
        description="A default category term used when source values are ambiguous, or unstated.",
    )
    minimum: Optional[Union[int, float]] = Field(
        None,
        description="An integer that specifies the minimum of a value, or the minimum number of characters of a string, depending on the field type.",
    )
    maximum: Optional[Union[int, float]] = Field(
        None,
        description="An integer that specifies the maximum of a value, or the maximum number of characters of a string, depending on the field type.",
    )

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True

    @validator("maximum")
    def is_maximum(cls, v, values, **kwargs):
        if "minimum" in values and values["minimum"] is not None and v is not None and v < values["minimum"]:
            raise ValueError(f"Maximum ({v}) must be greater than Minimum ({values['minimum']}).")
        return v

    @validator("category")
    def check_categories_unique(cls, v):
        category_names = []
        for c in v:
            category_names.append(c.name)
        if len(category_names) != len(set(category_names)):
            raise ValueError(
                f"Categories must be unique. There are {len(category_names) - len(set(category_names))} duplications."
            )
        return v



from enum import Enum


class MimeType(str, Enum):
    # Note that neither `feather` nor `parquet` yet have official mime types.
    CSV = "text/csv"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    PARQUET = "application/vnd.apache.parquet"
    PRQ = "application/vnd.apache.parquet"
    FEATHER = "application/vnd.apache.feather"
    FTR = "application/vnd.apache.feather"

    @classmethod
    def value_of(cls, value):
        # https://stackoverflow.com/a/56567247/295606
        # Allows:
        #   `MimeType.value_of("prq")`
        #   <MimeType.PARQUET: 'application/vnd.apache.parquet'>
        for k, v in cls.__members__.items():
            if k.upper() == value.upper():
                return v
        else:
            raise ValueError(f"'{cls.__name__}' enum not found for '{value}'")



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from whyqd.crosswalk.base import BaseMorphAction

if TYPE_CHECKING:
    import modin.pandas as pd
    from whyqd.models import FieldModel


class Action(BaseMorphAction):
    """Separate the string values in a single column into any number of new columns on a specified key.

    !!! tip "Script template"
        ```python
        "SEPARATE > ['destination_field_1', 'destination_field_2', etc. ] < 'by'::['source_field']"
        ```

        Where `'by'` can be any `string` with one **MAJOR** caveat. Since `::` is used in the query, you **cannot** use
        it for separating.

    !!! example
        ```python
        "SEPARATE > ['new_1', 'new_2', 'new_3', 'new_4'] < ';;'::['separate_column']"
        ```

        Will transform:

        | ID | separate_column         |
        |:---|:------------------------|
        | 2  | Dogs;;Cats;;Fish        |
        | 3  | Cats;;Bats;;Hats;;Mats  |
        | 4  | Sharks;;Crabs           |

        Bitwise into each destination field. If more fields are required than are assigned, will raise an error.

        | ID | separate_column         | new_1 | new_2 | new_3 | new_4 |
        |:---|:------------------------|:------|:------|:------|:------|
        | 2  | Dogs;;Cats;;Fish        |Dogs   | Cats  | Fish  |       |
        | 3  | Cats;;Bats;;Hats;;Mats  |Cats   | Bats  | Hats  | Mats  |
        | 4  | Sharks;;Crabs           |Sharks | Crabs |       |       |

        You will need a different strategy if separation results in more categories than you have destination fields for.
    """

    def __init__(self):
        self.name = "SEPARATE"
        self.title = "Separate strings"
        self.description = (
            "Separate the string values in a single column into any number of new columns on a specified key."
        )
        self.structure = ["fields"]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: list[FieldModel],
        source: FieldModel,
        source_param: str | None = None,
        rows: None = None,
        destination_param: None = None,
    ) -> pd.DataFrame:
        # separateting indefinite numbers of items
        # https://stackoverflow.com/a/50459986/295606
        # https://stackoverflow.com/a/39358924/295606
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.str.extract.html
        if not source_param:
            raise ValueError("SEPARATE action did not receive an expected separate-by parameter.")
        if len(source) != 1:
            raise ValueError("SEPARATE action should have only a single source field for separating.")
        separate_array = np.array([x.split(source_param) for x in df[source[0].name].array.ravel()], dtype=object)
        num_columns = max(map(len, separate_array))
        if num_columns == 1:
            return df
        if not isinstance(destination, list) or num_columns != len(destination):
            raise ValueError(
                f"SEPARATE action needs to separate {num_columns} columns by received {len(destination)} for destination."
            )
        new_columns = [c.name for c in destination]
        df[new_columns] = df[source[0].name].str.split(source_param, expand=True)
        return df



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseSchemaAction

if TYPE_CHECKING:
    import modin.pandas as pd
    from whyqd.models import FieldModel


class Action(BaseSchemaAction):
    """
    Add a new field to the dataframe, populated with a single value.

    !!! tip "Script template"
        ```python
        "NEW > 'destination_field' < ['value']"
        ```

        Where `value` is a unique `string` value which will be assigned to the entire column. This is useful where data form
        part of a series and each transformed dataset will be concatenated into a single larger table. This can function as
        a grouping identifier.

        **NOTE:** the `value` will be assigned as the default term for the `destination_field` FieldModel is part of the
        Schema, and the `default` in the `constraints`. Useful when creating a data series column. Will usually be added
        automatically after parsing of the scripts.

    !!! example
        ```python
        "NEW > 'la_code' < ['E06000044']"
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "NEW"
        self.title = "New"
        self.description = "Create a new field and assign a set value."
        self.structure = ["value"]

    def transform(self, *, df: pd.DataFrame, destination: FieldModel) -> pd.DataFrame:
        df[destination.name] = destination.constraints.default.name
        return df



from pathlib import Path

from whyqd.core.schema import SchemaDefinition
from whyqd.core.datasource import DataSourceDefinition
from whyqd.core.crosswalk import CrosswalkDefinition
from whyqd.core.transform import TransformDefinition

__version__ = (Path(__file__).resolve().parent / "VERSION").read_text()



from __future__ import annotations
from typing import List, Union, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    import modin.pandas as pd
    from whyqd.models import ModifierModel, FieldModel, SchemaActionModel, ColumnModel


class BaseSchemaAction:
    """Actions inherit from this base class which describes the core functions and methodology for an Action.

    Actions should redefine `name`, `title`, `description`, `modifiers` and `structure`, as well as produce a
    `transform` function. Everything else will probably remain as defined, but particularly complex Actions should
    modify as required.

    `structure` can be an empty list, but an Action may be defined by these parameters:

    * `modifier`: modifiers, of type `ModifierModel` defined by the ACTION and defining a `transform`.
    * `field`: the specific columns effected by the morph, a `list` of `ColumnModel` or, rarely, `FieldModel`.

    A standard script is::

        "ACTION > 'destination_field' < [modifier 'source_field', modifier 'source_field']"

    Where the structure of the source array is defined by the ACTION.
    """

    def __init__(self) -> None:
        from whyqd.parsers import CoreParser, DataSourceParser

        self.reader = DataSourceParser()
        self.core = CoreParser()
        self.name = ""
        self.title = ""
        self.description = ""
        # `structure` defines the format in which an action is written, and validated
        # can be - typically - any of `ColumnModel`, `ModifierModel`
        # additional terms will require overriding the `has_valid_structure` function
        self.structure = []

    @property
    def modifiers(self) -> Union[None, List[ModifierModel]]:
        """
        Describes the ModifierModels for the Action. Typical modifiers are `+` or `-` but the Action can implement any
        type of modifier as part of the `transform` function.

        As an example::

            [
                ModifierModel(**{
                    "name": "+",
                    "title": "Add",
                    "type": "modifier"
                }),
                ModifierModel(**{
                    "name": "-",
                    "title": "Subtract",
                    "type": "modifier"
                })
            ]

        Returns
        -------
        None or ModifierModel
            ModifierModel representation of the modifiers.
        """
        return None

    @property
    def modifier_terms(self) -> list[str]:
        if self.modifiers:
            return [m.name for m in self.modifiers]
        return []

    def get_modifier(self, *, term: str) -> Union[ModifierModel, None]:
        """Return a specific set of Modifier definitions in response to an Modifier name.

        Parameters
        ----------
        term: str
            A Modifier name.

        Returns
        -------
        ModifierModel, or None
            For the requested Modifier name. Or None, if it doesn't exist.
        """
        return next((m for m in self.modifiers if m.name == term), None)

    @property
    def settings(self) -> SchemaActionModel:
        """
        Returns the dict representation of the Action.

        Returns
        -------
        dict
            Dict representation of an Action.
        """
        from whyqd.models import SchemaActionModel

        action_settings = {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "structure": self.structure,
        }
        if self.modifiers:
            action_settings["modifiers"] = self.modifiers
        return SchemaActionModel(**action_settings)

    def validate(self, *, destination: FieldModel, source: list) -> bool:
        """
        Validate that script source structure conforms to the ACTION structure.

        Returns
        -------
        bool
        """
        if not isinstance(source, list):
            raise ValueError(f"Action source script does not conform to required structure. ({source})")
        for chunk in self.core.chunks(lst=source, n=len(self.structure)):
            # Loops through the phrasing of the structure, and checks that each term is as expected
            # e.g. [ModifierModel, FieldModel] for [modifier1, field1, modifier2, field2]
            # does not check that the actual terms match, though
            for i, term in enumerate(self.structure):
                if isinstance(chunk[i], term):
                    continue
                elif isinstance(chunk[i], dict):
                    # Nested source
                    if not chunk[i].get("action") and not chunk[i].get("source"):
                        raise ValueError(f"Nested script does not conform to required structure. ({chunk[i]})")
                    chunk[i]["action"].validate(source=chunk[i]["source"])
                else:
                    raise ValueError(
                        f"Source structure ({chunk}) doesn't conform to ACTION structure requirements ({self.structure})."
                    )
        return True

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: Optional[Union[FieldModel, ColumnModel]] = None,
        source: Optional[List[Union[ColumnModel, ModifierModel]]] = None,
    ) -> pd.DataFrame:
        """
        Perform a transformation. This function must be overridden by child Actions and describe a unique
        new method.

        .. warning:: Assumes that is a valid call. Will raise exceptions from inside Pandas if not.

        Parameters
        ----------
        df: DataFrame
            Working data to be transformed
        destination: FieldModel or ColumnModel, default None
            Destination column for the result of the Action. If required.
        source: list of ColumnModel and / or ModifierModel
            List of source columns and modifiers for the action. If required.

        Returns
        -------
        Dataframe
            Containing the implementation of the Action
        """
        return df



from __future__ import annotations
from pydantic import BaseModel, Field, validator
from uuid import UUID, uuid4

from whyqd.dtypes import FieldType


class ColumnModel(BaseModel):
    """Pandas DataFrame column model for maintaining source reference. Necessary to perform term matching in
    parsing operations (there can be some *really* fruity column names."""

    uuid: UUID = Field(default_factory=uuid4, description="Unique identity for the column. Automatically generated.")
    name: str = Field(
        ...,
        description="Column label / name from source data. In whatever format it was written, including leading/lagging spaces, non-printing characters, etc.",
    )
    dtype: FieldType = Field(
        default=FieldType.STRING,
        alias="type",
        description="A column must contain values of a specific type. Default is 'string'.",
    )

    class Config:
        use_enum_values = True
        validate_assignment = True
        allow_population_by_field_name = True

    @validator("dtype", pre=True, always=True)
    def generate_dtype(cls, v):
        if v in ["float64", "int64", "Float64", "Int64", FieldType.NUMBER.value]:
            return FieldType.NUMBER
        if v in ["datetime64[ns]", FieldType.DATETIME.value]:
            return FieldType.DATETIME
        return FieldType.STRING
